// display

// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////
// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 1
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "display"
#define LPSERVICE_1 80

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(25) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(5) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 3000 // the timeout during a GET, 3000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_AP_C_CONN_PERIOD 30000 // each 30 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 20000 // each 20 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD (1000*60*30) // each 30 minutes to find internet update; if cascade twice(no, use count) will become 15hr
#define TMO_MDNS_Q_PERIOD 60000 // each 60 seconds to query services
#define TMO_PING_PERIOD 6000 // each 3 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 0 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
#define USE_MDNS_STANDALONE // no auto hostname
//#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
//#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"display", 80},
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"display", 80},
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
};
#endif // if 1


// ************************** IoT END .h ******************************
////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_


#include <U8g2lib.h>
#include"Esp8266MultiPwms.h"
#include"Esp8266Vars2File.h"


#define VERSION_202104130001 "[[202104130001][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][WeMos TTGO][115200:160:26:4M:3MFS:512KSK:80M]]" // so the version number is the 3rd byte, length 12.


String web_link, ap_mode_speed_str;


// ------------------------------------ u8g2 config
#define OLED_SDA  2
#define OLED_SCL 14
#define OLED_RST  4
U8G2_SSD1306_128X32_UNIVISION_F_SW_I2C u8g2(U8G2_R0, OLED_SCL, OLED_SDA , OLED_RST);
const char *text = "U8g2";
char str1[17];
unsigned dir1=0;

// a null-terminated string already in an frmsz-size frm-frame with or without leading spaces will be wiggling in it.
// dir stores 0(forward) or 1(backward, non-zero) represents current direction which will be auto reversed.
void TextInFrameWiggle(unsigned &dir, byte frmsz, char *frm){
  
  for (int i=0, j=0; i<frmsz; i++){
    if (frm[i]!=' '){ // find the 1st non-space character incl. 0
      if (dir){ // backward
        if (i){
          do {frm[i-1]=frm[i];} while (frm[i++]);
          return;
        }
        else dir=0;
      }

      // forward
      j=i;
      while (frm[j++]);
      if (j-i==frmsz) return; // means no-leading-space string occupies entire frame
      if (j==frmsz){ // time to turnaround
        --i;
        dir=1;
        continue;
      }
      do {frm[j]=frm[j-1];} while (--j!=i);
      frm[j]=' ';
      return;
    }
  }
}

void setup_u8g2() {
  u8g2.begin();
  strcpy(str1, text);
}

void loop_u8g2_1(){
    TextInFrameWiggle(dir1, 17, str1);
    u8g2.drawStr(0,10,str1);
}

void loop_u8g2_2(){
    static String pr="";
    if (http_got_str!="" && pr!=http_got_str){
        pr=http_got_str;
    }

#if AP_MODE
    pr=ap_mode_speed_str;
#endif

    if (pr!="") u8g2.drawStr(0,28,pr.c_str());

}


// ------------------------------------ for periodically u8g2 drawing
#define DRAW_INT_GPIO           12
#define DRAW_TRG_GPIO           13
#define DRAW_PERIOD             100000 // 100ms. less than 100ms will wdt-reset.

typedef struct sFooLinkedList{
    void (*fp)();
    sFooLinkedList *next;
    sFooLinkedList(): fp(0), next(0){;};
    ~sFooLinkedList(){delete this->next;};
} sFooLinkedList;

cHwTimer timer_draw_manip; // be aware that hwtimer1 must bring up first.
IRAM_ATTR unsigned timer_trigger=0;
IRAM_ATTR void isr_draw_manip_timer(){digitalWrite(DRAW_TRG_GPIO, timer_trigger^=1);}
IRAM_ATTR unsigned DrawSubmit=0; // set 1 to start drawing
IRAM_ATTR unsigned AutoDrawSubmit=0; // set 1 to start auto drawing
sFooLinkedList *draw_link=0;

void DrawSetup(){
    pinMode(DRAW_TRG_GPIO, OUTPUT); // this gpio is to trigger intr gpio
    digitalWrite(DRAW_TRG_GPIO, LOW);
    attachInterrupt(digitalPinToInterrupt(DRAW_INT_GPIO), isr_draw_manip, CHANGE);

    timer_draw_manip.setTimer(DRAW_PERIOD, isr_draw_manip_timer);
}

void DrawConfig(void (*fptr)()){ // insert function pointers to which draw the buffer.
    if (!draw_link){
        draw_link=new sFooLinkedList;
        draw_link->fp=fptr;
        return;
    }
    sFooLinkedList *a_link=draw_link;
    while (a_link->next) a_link=a_link->next;
    a_link->next=new sFooLinkedList;
    a_link->next->fp=fptr;
}

ICACHE_RAM_ATTR void isr_draw_manip(){
    if (DrawSubmit==1){
        DrawSubmit=2;
        u8g2.clearBuffer();
        u8g2.setFont(u8g2_font_8x13B_mf);
        sFooLinkedList *a_link=draw_link;
        while (a_link){
            if (a_link->fp) (a_link->fp)();
            a_link=a_link->next;
        }
        u8g2.sendBuffer();
        DrawSubmit=0;
    }
    if (AutoDrawSubmit) DrawSubmit=1;
}

void DrawEnd(){ // must called if do not draw anymore
    delete draw_link;
}


// ------------------------------------ http
void my_http_request(){ // called only if http_request_ip assigned.
    static unsigned tmo=micros();
    if (Timeout(433, tmo)){ // each 0.433 seconds to send http request.
        HTTPC_http_client_connect(web_link.c_str());
    }
}

void my_setup_http_server(){

    uri_list+=\
        "/version\r\n"\
        "/s (for ap mode only)\r\n";

    server_p->on("/version", []() { // show version
        server_p->send(200, "text/plain", VERSION_202104130001);
    });

#if AP_MODE
    server_p->on("/s", []() {
        int i=(random(3)? -1: 1);
        static int j=0;
        static unsigned k=0;
        if (!k && random(17)>14) k=2;
        else if (k){k--; i=1;}
        j+=i;
        if (j<0) j=0;
        else if (j>7) j=7;
        ap_mode_speed_str=(j*1000+random(1000));
        server_p->send(200, "text/plain", ap_mode_speed_str.c_str());
    });
#endif
}


void init_params(){
    cDataBase s("/settings");
    s.Open();
    unsigned BootTimes=0;

    if (s.ReadWords(BootTimes, "BootTimes", 0)){
        printf("load params success: %s %u\r\n", "BootTimes", BootTimes);
    }

    s.Close();
    s.ReplaceWordsEx("BootTimes", 0, ++BootTimes);
    if (!(BootTimes%5)) s.Renew();
}


void setup(){
    Serial.begin(115200);
    delay(5000);

    init_params();

    randomSeed(micros()%ESP.getChipId());

    setup_IoT();
    setup_u8g2();
    DrawSetup();
    DrawConfig(loop_u8g2_1);
    DrawConfig(loop_u8g2_2);
}

void loop(){
    static unsigned x=1;
    if (x && *remote_ap_ip){
        if (!*http_request_ip) strcpy(http_request_ip, remote_ap_ip);
        web_link=String("http://")+http_request_ip+"/s";
        x=0;
    }
    loop_IoT();

    static unsigned tmo=micros();
    if (Timeout(10000, tmo, true)){
        AutoDrawSubmit=1;
    }
}
